==> v0.0.5.zip: ag/templates/_main.go <==
package main

import (
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"regexp"
	"strings"
	"syscall"
	"time"

	"apigo.cc/gojs"
	{{range $pkg, $ver := .Module}}
	_ "{{$pkg}}"
	{{- end}}
	{{range $pkg, $ver := .ExtraImport}}
	_ "{{$pkg}}"
	{{- end}}
	"github.com/ssgo/log"
	"github.com/ssgo/u"
)

var _mainFile = "{{.Main}}"
var appName = "{{.Name}}"
var appVersion = "{{.Version}}"
var watchPath = []string{"."}
var watchFileType = []string{"js", "ts", "yml", "yaml", "json"}
var watchWeb = []string{}
var buildWebFrom = ""
var buildWebTo = ""
var buildWebCopyFiles = map[string]string{}

func init() {
	{{- range $alias, $pkg := .ModuleAlias }}
	gojs.Alias("{{$alias}}", "{{$pkg}}")
	{{- end}}
	{{- if .WatchPath }}
	watchPath = u.AppendUniqueStrings(watchPath, u.SplitWithoutNone("{{.WatchPath}}", ","))
	{{- end }}
	{{- if .WatchFileType }}
	watchFileType = u.AppendUniqueStrings(watchFileType, u.SplitWithoutNone("{{.WatchFileType}}", ","))
	{{- end }}

	{{- if .WatchWeb }}
	watchWeb = u.SplitWithoutNone("{{.WatchWeb}}", ",")
	{{- end }}

	{{- if .BuildWeb }}
	buildWebFrom = "{{.BuildWeb.From}}"
	buildWebTo = "{{.BuildWeb.To}}"
	if buildWebTo == "" {
		buildWebTo = "www"
	}
	    {{- range $k, $v := .BuildWeb.CopyFiles}}
	buildWeb.CopyFiles["{{$k}}"] = "{{$v}}"
	    {{- end }}
	watchWeb = u.AppendUniqueString(watchWeb, buildWebTo)
	{{- end }}

	if len(watchWeb) > 0 {
		watchPath = u.AppendUniqueStrings(watchPath, watchWeb)
		watchFileType = u.AppendUniqueStrings(watchFileType, u.SplitWithoutNone("html,css,png,jpg,jpeg,gif,webp,svg,scss,less,ejs,mjs,woff,woff2,ttf,ico", ","))
		if len(watchWeb) > 0 {
			for i, v := range watchWeb {
				watchWeb[i], _ = filepath.Abs(v)
			}
		}	
	}
}

var idFixMatcher = regexp.MustCompile(`[^a-zA-Z0-9_]`)
var testCaseOnStartMatcher = regexp.MustCompile(`(?m)^\s*function\s+onStart\s*\(`)
var testCaseOnStopMatcher = regexp.MustCompile(`(?m)^\s*function\s+onStop\s*\(`)
var testCaseMatcher = regexp.MustCompile(`(?m)^\s*function\s+test_(\w+)\s*\(`)

func setSSKey(key, iv []byte) {
	gojs.SetSSKey(key, iv)
}

func main() {
	if appName == "" {
		appName = filepath.Base(os.Args[0])
		appName = strings.TrimSuffix(appName, ".exe")
	}

	if appVersion == "" {
		appVersion = "0.0.1"
	}
	cmd := ""
	id := ""
	mainFile := _mainFile
	isWatch := false
	startArgs := make([]string, 0)
	mainArgs := make([]any, 0)
	for i := 1; i < len(os.Args); i++ {
		arg := os.Args[i]
		switch arg {
		case "--help", "-help", "-h":
			printUsage()
			return
		case "--version", "-version", "-v":
			fmt.Println(appName, appVersion)
			return
		case "--export", "-export":
			fmt.Println(u.Cyan(gojs.ExportForDev()))
			return
		case "start", "stop", "restart", "reload", "test":
			if len(mainArgs) == 0 && cmd == "" {
				cmd = arg
			} else {
				mainArgs = append(mainArgs, arg)
			}
		case "--env", "-env", "-e":
			i++
			a := strings.SplitN(os.Args[i], "=", 2)
			os.Setenv(a[0], a[1])
		case "-id":
			i++
			id = strings.TrimSpace(os.Args[i])
		case "-main":
			i++
			mainFile = strings.TrimSpace(os.Args[i])
			startArgs = append(startArgs, "-main", mainFile)
		case "-watch", "-w":
			isWatch = true
			startArgs = append(startArgs, arg)
		default:
			mainArgs = append(mainArgs, arg)
			startArgs = append(startArgs, arg)
		}
	}

	if cmd != "test" {
		if !u.FileExists(mainFile) {
			log.DefaultLogger.Error("main file not found", "mainFile", mainFile)
			return
		}
		if id == "" {
			id = fmt.Sprintf("%s_%s_%s", appName, mainFile, appVersion)
		}
		id = idFixMatcher.ReplaceAllString(id, "_")
	}

	homeDir, _ := os.UserHomeDir()
	pidFile := filepath.Join(homeDir, ".pids", id+".pid")
	pid := u.Int(u.ReadFileN(pidFile))
	switch cmd {
	case "start":
		if pid > 0 && processIsExists(pid) {
			log.DefaultLogger.Info("process already started", "pid", pid)
		} else {
			startProcess(pidFile, startArgs)
		}
	case "stop":
		if pid > 0 {
			killProcess(pid, mainFile)
			_ = os.Remove(pidFile)
		}
	case "reload":
		if pid > 0 {
			kill(pid, syscall.Signal(0x1e))
		}
	case "restart":
		if pid > 0 {
			killProcess(pid, mainFile)
			_ = os.Remove(pidFile)
		}
		startProcess(pidFile, startArgs)
	case "test":
		testPath := "."
		if u.FileExists("tests") {
			testPath = "tests"
		}
		if isWatch {
			var wr *gojs.WatchRunner
			wr, _ = gojs.Watch(watchPath, watchFileType, func(changes []string) {
				if len(changes) == 1 && strings.HasSuffix(changes[0], "_test.js") {
					runTest(changes[0], wr)
				} else {
					runAllTest(testPath, wr)
				}
			})
		} else {
			runAllTest(testPath, nil)
		}
		gojs.WaitAll()
	default:
		defer os.Remove(pidFile)
		if isWatch {
			var wr *gojs.WatchRunner
			var rt *gojs.Runtime
			buildWebFromPath := ""
			if buildWebFrom != "" {
				buildWebFromPath, _ = filepath.Abs(buildWebFrom)
			}
			wr, _ = gojs.Watch(watchPath, watchFileType, func(changes []string) {
				buildWeb := false
				reloadWeb := false
				restartApp := false
				var onWatch func(string)
				if len(changes) == 0 {
					buildWeb = buildWebFromPath != ""
					restartApp = true
				} else {
					if len(watchWeb) > 0 {
						if rt != nil {
							if onWatch1 := rt.GetGoData("onWatch"); onWatch1 != nil {
								if onWatch2, ok := onWatch1.(func(string)); ok {
									onWatch = onWatch2
								}
							}
							if onWatch != nil {
								for _, changedFile := range changes {
									if buildWebFromPath != "" && strings.HasPrefix(changedFile, buildWebFromPath) {
										buildWeb = true
									} else {
										inWeb := false
										for _, watchWebPath := range watchWeb {
											if strings.HasPrefix(changedFile, watchWebPath) {
												inWeb = true
												reloadWeb = true
											}
										}
										if !inWeb {
											restartApp = true
										}
									}
								}
							}
						}
					}
				}

				if restartApp {
					fmt.Print(u.BMagenta("restartApp"))
				}
				if buildWeb {
					fmt.Print(u.BMagenta("buildWeb"))
				}
				if buildWebFromPath == "" && reloadWeb {
					fmt.Print(u.BMagenta("reloadWeb"))
				}
				fmt.Println()
				if restartApp {
					gojs.StopForWatch()

					// 启动新的程序
					rt = gojs.New()
					rt.SetGoData("inWatch", true)
					wr.SetModuleLoader(rt)
					err := rt.StartFromFile(mainFile)
					if err != nil {
						fmt.Println(u.BRed(err.Error()))
						fmt.Println(u.Red("  " + strings.Join(rt.GetCallStack(), "\n  ")))
						return
					}
					result, err := rt.RunMain(mainArgs...)
					if err != nil {
						fmt.Println(u.BRed(err.Error()))
						fmt.Println(u.Red("  " + strings.Join(rt.GetCallStack(), "\n  ")))
					} else if result != nil {
						fmt.Println(u.Cyan(u.JsonP(result)))
					}
				}

				if buildWeb {
					cmd := exec.Command("ag", "build", "web")
					cmd.Stdin = os.Stdin
					cmd.Stdout = os.Stdout
					cmd.Stderr = os.Stderr
					cmd.Run()
					if onWatch != nil {
						onWatch(strings.Join(changes, "\n"))
					}
				}

				if buildWebFromPath == "" && reloadWeb {
					onWatch(strings.Join(changes, "\n"))
				}
			})
		} else {
			if r, err := gojs.RunFile(mainFile, mainArgs...); err != nil {
				log.DefaultLogger.Error("run failed for "+mainFile, "err", err.Error())
			} else if r != nil {
				log.DefaultLogger.Error("run "+mainFile, "result", r)
			}
		}
		gojs.WaitAll()
	}
}

func runAllTest(testPath string, wr *gojs.WatchRunner) {
	for _, f := range u.ReadDirN(testPath) {
		if !f.IsDir && strings.HasSuffix(f.Name, "_test.js") {
			runTest(f.Name, wr)
		}
	}
}

func runTest(filename string, wr *gojs.WatchRunner) {
	vm := gojs.New()
	if wr != nil {
		wr.SetModuleLoader(vm)
	}
	if _, err := vm.RunFile(filename); err != nil {
		fmt.Println(u.BRed("run "+filename), u.Red(err.Error()))
		return
	}
	code := u.ReadFileN(filename)
	if testCaseOnStartMatcher.MatchString(code) {
		if _, err := vm.RunCode(`onStart()`); err != nil {
			fmt.Println(u.Red(filename), u.BRed("onStart"), u.Red(err.Error()))
			return
		} else {
			fmt.Println(u.Green(filename), u.BGreen("onStart"), u.Green("success"))
		}
	}
	for _, m := range testCaseMatcher.FindAllStringSubmatch(code, 1000) {
		testCaseName := m[1]
		if r, err := vm.RunCode(testCaseName + "()"); err != nil {
			fmt.Println(u.Red(filename), u.BRed(testCaseName), u.Red(err.Error()))
		} else if r == true {
			fmt.Println(u.Green(filename), u.BGreen(testCaseName), u.Green("pass"))
		} else {
			fmt.Println(u.Red(filename), u.BRed(testCaseName), u.Red(u.StringP(r)))
		}
	}
	if testCaseOnStopMatcher.MatchString(code) {
		if _, err := vm.RunCode(`onStop()`); err != nil {
			fmt.Println(u.Red(filename), u.BRed("onStop"), u.Red(err.Error()))
		} else {
			fmt.Println(u.Green(filename), u.BGreen("onStop"), u.Green("success"))
		}
	}
}

func startProcess(pidFile string, startArgs []string) {
	cmd := exec.Command(os.Args[0], startArgs...)
	// cmd.Stdout = os.Stdout
	// cmd.Stderr = os.Stderr
	if err := cmd.Start(); err == nil {
		u.WriteFile(pidFile, u.String(cmd.Process.Pid))
		log.DefaultLogger.Info("started", "appName", appName, "appVersion", appVersion, "startArgs", startArgs, "pid", cmd.Process.Pid)
	} else {
		log.DefaultLogger.Error("start failed", "appName", appName, "appVersion", appVersion, "startArgs", startArgs, "err", err)
	}
	os.Exit(0)
}

func kill(pid int, sig os.Signal) error {
	if proc, err := os.FindProcess(pid); err == nil {
		return proc.Signal(sig)
	} else {
		return err
	}
}

func killProcess(pid int, mainFile string) {
	if err := kill(pid, syscall.SIGTERM); err == nil {
		log.DefaultLogger.Info("killing", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid)
		t1 := time.Now().UnixMilli()
		for i := 0; i < 100; i++ {
			if !processIsExists(pid) {
				log.DefaultLogger.Info("killed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid, "usedTime", (time.Duration(time.Now().UnixMilli()-t1) * time.Millisecond).String())
				return
			}
			time.Sleep(100 * time.Millisecond)
		}
		err := kill(pid, syscall.SIGKILL)
		if processIsExists(pid) {
			log.DefaultLogger.Error("fource kill failed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid, "err", err)
		} else {
			log.DefaultLogger.Info("fource killed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid)
		}
	}
}

func processIsExists(pid int) bool {
	if proc, err := os.FindProcess(pid); err == nil {
		if err2 := proc.Signal(syscall.Signal(0)); err2 == nil {
			return true
		}
	}
	return false
}

func printUsage() {
	fmt.Println(appName, appVersion)
	fmt.Println("Usage:")
	fmt.Println("  ", appName, "[-main mainFile] ...", "run script")
	fmt.Println("  ", appName, "-version|-v", "show version")
	fmt.Println("  ", appName, "-export", "export for development")
	fmt.Println("  ", appName, "-env|-e", "set env")
	fmt.Println("  ", appName, "start [-id id] [-main mainFile]", "start server")
	fmt.Println("  ", appName, "stop [-id id]", "stop server")
	fmt.Println("  ", appName, "restart [-id id] [-main mainFile]", "restart server")
	fmt.Println("  ", appName, "reload [-id id] [-main mainFile]", "reload config")
	fmt.Println("  ", appName, "test", "test *_test.js")
	fmt.Println("  ", appName, "help", "- show help")
}

==> v0.0.6.zip: ag/templates/_main.go <==
package main

import (
	"fmt"
	"os"
	"os/exec"
	"path/filepath"
	"regexp"
	"strings"
	"syscall"
	"time"

	"apigo.cc/gojs"
	{{range $pkg, $ver := .Module}}
	_ "{{$pkg}}"
	{{- end}}
	{{range $pkg, $ver := .ExtraImport}}
	_ "{{$pkg}}"
	{{- end}}
	"github.com/ssgo/log"
	"github.com/ssgo/u"
)

var _mainFile = "{{.Main}}"
var appName = "{{.Name}}"
var appVersion = "{{.Version}}"
var watchPath = []string{"."}
var watchFileType = []string{"js", "ts", "yml", "yaml", "json"}
var watchWeb = []string{}
var buildWebFrom = ""
var buildWebTo = ""
var buildWebCopyFiles = map[string]string{}

func init() {
	{{- range $alias, $pkg := .ModuleAlias }}
	gojs.Alias("{{$alias}}", "{{$pkg}}")
	{{- end}}
	{{- if .WatchPath }}
	watchPath = u.AppendUniqueStrings(watchPath, u.SplitWithoutNone("{{.WatchPath}}", ","))
	{{- end }}
	{{- if .WatchFileType }}
	watchFileType = u.AppendUniqueStrings(watchFileType, u.SplitWithoutNone("{{.WatchFileType}}", ","))
	{{- end }}

	{{- if .WatchWeb }}
	watchWeb = u.SplitWithoutNone("{{.WatchWeb}}", ",")
	{{- end }}

	{{- if .BuildWeb }}
	buildWebFrom = "{{.BuildWeb.From}}"
	buildWebTo = "{{.BuildWeb.To}}"
	if buildWebTo == "" {
		buildWebTo = "www"
	}
	    {{- range $k, $v := .BuildWeb.CopyFiles}}
	buildWeb.CopyFiles["{{$k}}"] = "{{$v}}"
	    {{- end }}
	watchWeb = u.AppendUniqueString(watchWeb, buildWebTo)
	{{- end }}

	if len(watchWeb) > 0 {
		watchPath = u.AppendUniqueStrings(watchPath, watchWeb)
		watchFileType = u.AppendUniqueStrings(watchFileType, u.SplitWithoutNone("html,css,png,jpg,jpeg,gif,webp,svg,scss,less,ejs,mjs,woff,woff2,ttf,ico", ","))
		if len(watchWeb) > 0 {
			for i, v := range watchWeb {
				watchWeb[i], _ = filepath.Abs(v)
			}
		}	
	}
}

var idFixMatcher = regexp.MustCompile(`[^a-zA-Z0-9_]`)
var testCaseOnStartMatcher = regexp.MustCompile(`(?m)^\s*function\s+onStart\s*\(`)
var testCaseOnStopMatcher = regexp.MustCompile(`(?m)^\s*function\s+onStop\s*\(`)
var testCaseMatcher = regexp.MustCompile(`(?m)^\s*function\s+test(\w+)\s*\(`)

func setSSKey(key, iv []byte) {
	gojs.SetSSKey(key, iv)
}

func main() {
	if appName == "" {
		appName = filepath.Base(os.Args[0])
		appName = strings.TrimSuffix(appName, ".exe")
	}

	if appVersion == "" {
		appVersion = "0.0.1"
	}
	cmd := ""
	id := ""
	mainFile := _mainFile
	isWatch := false
	startArgs := make([]string, 0)
	mainArgs := make([]any, 0)
	for i := 1; i < len(os.Args); i++ {
		arg := os.Args[i]
		switch arg {
		case "--help", "-help", "-h":
			printUsage()
			return
		case "--version", "-version", "-v":
			fmt.Println(appName, appVersion)
			return
		case "--export", "-export":
			fmt.Println(u.Cyan(gojs.ExportForDev()))
			return
		case "start", "stop", "restart", "reload", "test":
			if len(mainArgs) == 0 && cmd == "" {
				cmd = arg
			} else {
				mainArgs = append(mainArgs, arg)
			}
		case "--env", "-env", "-e":
			i++
			a := strings.SplitN(os.Args[i], "=", 2)
			os.Setenv(a[0], a[1])
		case "-id":
			i++
			id = strings.TrimSpace(os.Args[i])
		case "-main":
			i++
			mainFile = strings.TrimSpace(os.Args[i])
			startArgs = append(startArgs, "-main", mainFile)
		case "-watch", "-w":
			isWatch = true
			startArgs = append(startArgs, arg)
		default:
			mainArgs = append(mainArgs, arg)
			startArgs = append(startArgs, arg)
		}
	}

	if cmd != "test" {
		if !u.FileExists(mainFile) {
			log.DefaultLogger.Error("main file not found", "mainFile", mainFile)
			return
		}
		if id == "" {
			id = fmt.Sprintf("%s_%s_%s", appName, mainFile, appVersion)
		}
		id = idFixMatcher.ReplaceAllString(id, "_")
	}

	homeDir, _ := os.UserHomeDir()
	pidFile := filepath.Join(homeDir, ".pids", id+".pid")
	pid := u.Int(u.ReadFileN(pidFile))
	switch cmd {
	case "start":
		if pid > 0 && processIsExists(pid) {
			log.DefaultLogger.Info("process already started", "pid", pid)
		} else {
			startProcess(pidFile, startArgs)
		}
	case "stop":
		if pid > 0 {
			killProcess(pid, mainFile)
			_ = os.Remove(pidFile)
		}
	case "reload":
		if pid > 0 {
			kill(pid, syscall.Signal(0x1e))
		}
	case "restart":
		if pid > 0 {
			killProcess(pid, mainFile)
			_ = os.Remove(pidFile)
		}
		startProcess(pidFile, startArgs)
	case "test":
		testPath := "."
		if u.FileExists("tests") {
			testPath = "tests"
		}
		if isWatch {
			var wr *gojs.WatchRunner
			wr, _ = gojs.Watch(watchPath, watchFileType, nil, func(changes []string) {
				if len(changes) == 1 && strings.HasSuffix(changes[0], "_test.js") {
					runTest(changes[0], wr)
				} else {
					runAllTest(testPath, wr)
				}
			})
		} else {
			runAllTest(testPath, nil)
		}
		gojs.WaitAll()
	default:
		defer os.Remove(pidFile)
		if isWatch {
			var wr *gojs.WatchRunner
			var rt *gojs.Runtime
			buildWebFromPath := ""
			if buildWebFrom != "" {
				buildWebFromPath, _ = filepath.Abs(buildWebFrom)
			}
			wr, _ = gojs.Watch(watchPath, watchFileType, watchWeb, func(changes []string) {
				buildWeb := false
				reloadWeb := false
				restartApp := false
				var onWatch func(string)
				if len(changes) == 0 {
					buildWeb = buildWebFromPath != ""
					restartApp = true
				} else {
					if len(watchWeb) > 0 {
						if rt != nil {
							if onWatch1 := rt.GetGoData("onWatch"); onWatch1 != nil {
								if onWatch2, ok := onWatch1.(func(string)); ok {
									onWatch = onWatch2
								}
							}
							// if onWatch != nil {
							for _, changedFile := range changes {
								if buildWebFromPath != "" && strings.HasPrefix(changedFile, buildWebFromPath) {
									buildWeb = true
								} else {
									inWeb := false
									for _, watchWebPath := range watchWeb {
										if strings.HasPrefix(changedFile, watchWebPath) {
											inWeb = true
											reloadWeb = true
										}
									}
									if !inWeb {
										restartApp = true
									}
								}
							}
						}
					} else {
						restartApp = true
					}
				}

				// fmt.Print(u.BMagenta("============"), restartApp, buildWeb, buildWebFromPath == "" && reloadWeb)
				if restartApp {
					fmt.Print(u.BMagenta("restartApp"))
				}
				if buildWeb {
					fmt.Print(u.BMagenta("buildWeb"))
				}
				if buildWebFromPath == "" && reloadWeb {
					fmt.Print(u.BMagenta("reloadWeb"))
				}
				fmt.Println()
				if restartApp {
					gojs.StopForWatch()

					// 启动新的程序
					rt = gojs.New()
					rt.SetGoData("inWatch", true)
					wr.SetModuleLoader(rt)
					err := rt.StartFromFile(mainFile)
					if err != nil {
						fmt.Println(u.BRed(err.Error()))
						fmt.Println(u.Red("  " + strings.Join(rt.GetCallStack(), "\n  ")))
						return
					}
					result, err := rt.RunMain(mainArgs...)
					if err != nil {
						fmt.Println(u.BRed(err.Error()))
						fmt.Println(u.Red("  " + strings.Join(rt.GetCallStack(), "\n  ")))
					} else if result != nil {
						fmt.Println(u.Cyan(u.JsonP(result)))
					}
				}

				if buildWeb {
					cmd := exec.Command("ag", "build", "web")
					cmd.Stdin = os.Stdin
					cmd.Stdout = os.Stdout
					cmd.Stderr = os.Stderr
					cmd.Run()
					if onWatch != nil {
						onWatch(strings.Join(changes, "\n"))
					}
				}

				if buildWebFromPath == "" && reloadWeb {
					onWatch(strings.Join(changes, "\n"))
				}
			})
		} else {
			if r, err := gojs.RunFile(mainFile, mainArgs...); err != nil {
				log.DefaultLogger.Error("run failed for "+mainFile, "err", err.Error())
			} else if r != nil {
				log.DefaultLogger.Error("run "+mainFile, "result", r)
			}
		}
		gojs.WaitAll()
	}
}

func runAllTest(testPath string, wr *gojs.WatchRunner) {
	for _, f := range u.ReadDirN(testPath) {
		if !f.IsDir && strings.HasSuffix(f.Name, "_test.js") {
			runTest(f.Name, wr)
		}
	}
}

func runTest(filename string, wr *gojs.WatchRunner) {
	vm := gojs.New()
	if wr != nil {
		wr.SetModuleLoader(vm)
	}
	if _, err := vm.RunFile(filename); err != nil {
		fmt.Println(u.BRed("run "+filename), u.Red(err.Error()))
		return
	}
	code := u.ReadFileN(filename)
	if testCaseOnStartMatcher.MatchString(code) {
		if _, err := vm.RunCode(`onStart()`); err != nil {
			fmt.Println(u.Red(filename), u.BRed("onStart"), u.Red(err.Error()))
			return
		} else {
			fmt.Println(u.Green(filename), u.BGreen("onStart"), u.Green("success"))
		}
	}
	for _, m := range testCaseMatcher.FindAllStringSubmatch(code, 1000) {
		testCaseName := m[1]
		if r, err := vm.RunCode(testCaseName + "()"); err != nil {
			fmt.Println(u.Red(filename), u.BRed(testCaseName), u.Red(err.Error()))
		} else if r == true {
			fmt.Println(u.Green(filename), u.BGreen(testCaseName), u.Green("pass"))
		} else {
			fmt.Println(u.Red(filename), u.BRed(testCaseName), u.Red(u.StringP(r)))
		}
	}
	if testCaseOnStopMatcher.MatchString(code) {
		if _, err := vm.RunCode(`onStop()`); err != nil {
			fmt.Println(u.Red(filename), u.BRed("onStop"), u.Red(err.Error()))
		} else {
			fmt.Println(u.Green(filename), u.BGreen("onStop"), u.Green("success"))
		}
	}
}

func startProcess(pidFile string, startArgs []string) {
	cmd := exec.Command(os.Args[0], startArgs...)
	// cmd.Stdout = os.Stdout
	// cmd.Stderr = os.Stderr
	if err := cmd.Start(); err == nil {
		u.WriteFile(pidFile, u.String(cmd.Process.Pid))
		log.DefaultLogger.Info("started", "appName", appName, "appVersion", appVersion, "startArgs", startArgs, "pid", cmd.Process.Pid)
	} else {
		log.DefaultLogger.Error("start failed", "appName", appName, "appVersion", appVersion, "startArgs", startArgs, "err", err)
	}
	os.Exit(0)
}

func kill(pid int, sig os.Signal) error {
	if proc, err := os.FindProcess(pid); err == nil {
		return proc.Signal(sig)
	} else {
		return err
	}
}

func killProcess(pid int, mainFile string) {
	if err := kill(pid, syscall.SIGTERM); err == nil {
		log.DefaultLogger.Info("killing", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid)
		t1 := time.Now().UnixMilli()
		for i := 0; i < 100; i++ {
			if !processIsExists(pid) {
				log.DefaultLogger.Info("killed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid, "usedTime", (time.Duration(time.Now().UnixMilli()-t1) * time.Millisecond).String())
				return
			}
			time.Sleep(100 * time.Millisecond)
		}
		err := kill(pid, syscall.SIGKILL)
		if processIsExists(pid) {
			log.DefaultLogger.Error("fource kill failed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid, "err", err)
		} else {
			log.DefaultLogger.Info("fource killed", "appName", appName, "appVersion", appVersion, "mainFile", mainFile, "pid", pid)
		}
	}
}

func processIsExists(pid int) bool {
	if proc, err := os.FindProcess(pid); err == nil {
		if err2 := proc.Signal(syscall.Signal(0)); err2 == nil {
			return true
		}
	}
	return false
}

func printUsage() {
	fmt.Println(appName, appVersion)
	fmt.Println("Usage:")
	fmt.Println("  ", appName, "[-main mainFile] ...", "run script")
	fmt.Println("  ", appName, "-version|-v", "show version")
	fmt.Println("  ", appName, "-export", "export for development")
	fmt.Println("  ", appName, "-env|-e", "set env")
	fmt.Println("  ", appName, "start [-id id] [-main mainFile]", "start server")
	fmt.Println("  ", appName, "stop [-id id]", "stop server")
	fmt.Println("  ", appName, "restart [-id id] [-main mainFile]", "restart server")
	fmt.Println("  ", appName, "reload [-id id] [-main mainFile]", "reload config")
	fmt.Println("  ", appName, "test", "test *_test.js")
	fmt.Println("  ", appName, "help", "- show help")
}
